'----------------------------------------------------------------------------------------
'    Name:  Sd_Card_Spi_I.Inc
'    Type:  Tiger-BASIC(tm) Source Code (SD Card Access)
' Release:  1.3
' Purpose:  SD Card Access - Implementation of SD Card direct access
'
'  (C) - Copyright Wilke Technology, P.O.Box 1727, D-52018 Aachen, Germany
'----------------------------------------------------------------------------------------
'
' Thank you for using BASIC Tiger(tm) in your products. If you have questions, ideas
' or special needs, please contact your next distributor or the Tiger support team
' and visit our web site:
'
'	Wilke Technology GmbH
'	The Tiger Support Team
'	P.O.Box 1727,       D-52018 Aachen, Germany
'	Krefelder Str. 147, D-52070 Aachen, Germany
'
'	email: support@wilke-technology.com  (English)
'	email: support@wilke.de              (German)
'	Phone: +49 (241) 918 900    Mo to Fr, 7:00 to 16:00 (GMT)
'	Fax:   +49 (241) 918 9068
'
' New information, new drivers and free downloads see:
'
'	www.wilke-technology.com	     (English)
'	www.wilke.de                         (German)
'
' Sincerely,
'
'	Your Tiger Support Team
'
'----------------------------------------------------------------------------------------


#ifndef __SD_CARD_SPI_I
#define __SD_CARD_SPI_I

'' Pin Definitions
#include sd_card_pins_d.inc

' ******************************************************************************** '
'                               GLOBAL VARIABLES                                   '
' ******************************************************************************** '

'' The maximum read data block length is computed as 2READ_BL_LEN.
'' The maximum block length might therefore be in the range 512...2048 bytes.
long lSDBlockLen

' ---------------------------------------------------------------------------------'
'           		         CSD Register                                          '	
' ---------------------------------------------------------------------------------'
'' Note that in the SD Card, the WRITE_BL_LEN is always equal to READ_BL_LEN.
'' This variables must be initialized bei Initialization.
byte bSDRegCSDReadBlLen

'' READ_BL_PARTIAL is always set to 1 in the SD Card.
'' Partial Block Read is always allowed in the SD Card.
'' It means that smaller blocks can be used as well. The minimum block size is one byte.
byte bSDRegCSDReadBlPartial

'' TAACDefines the asynchronous part (relative to the SD Card clock (CLK))
'' of the read access time.
byte bSDRegCSDTaac

'' NSACDefines the worst case for the clock dependent factor of the data access time.
'' The unit for NSAC is 100 clock cycles. Therefore, the maximal value for the clock
'' dependent part of the read access time is 25.5k clock cycles.
byte bSDRegCSDNsac

'' C_SIZE (Device Size)This parameter is used to compute the card capacity
'' (does not include security protected area).
word wSDRegCSDCsize

'' C_SIZE_MULT (Device Size Multiplier)This parameter is used for coding a factor MULT for computing the
'' total device size (see C_SIZE). The factor MULT is defined as 2C_SIZE_MULT+2.
byte bSDRegCSDCsizeMult

'' SECTOR_SIZEThe size of an erasable sector. The contents of this register is a 7-bit binary coded value,
'' defining the number of write blocks (see WRITE_BL_LEN). The actual size is computed by increasing this number
'' by one. A value of zero means 1 write block, 127 means 128 blocks.
byte bSDRegCSDSectorSize

'' WP_GRP_SIZEThe size of a write protected group. The contents of this register is a 7-bit binary coded value,
'' defining the number of Erase Groups (see SECTOR_SIZE). The actual size is computed by increasing this number
'' by one. A value of zero means 1 erase group, 127 means 128 erase groups.
byte bSDRegCSDWpGrpSize

'' WRITE_BL_LENThe maximum write data block length is computed as 2WRITE_BL_LEN. The maximum block
'' length might therefore be in the range from 512 up to 2048 bytes. A Write Block Length of 512 bytes is always
'' supported. Note that in the SD Card, the WRITE_BL_LEN is always equal to READ_BL_LEN.
byte bSDRegCSDWriteBlLen

'' WRITE_BL_PARTIALDefines whether partial block sizes can be used in block write commands.
'' WRITE_BL_PARTIAL=0 means that only the WRITE_BL_LEN block size, and its partial derivatives
'' in resolution of units of 512 blocks, can be used for block oriented data write.
'' WRITE_BL_PARTIAL=1 means that smaller blocks can be used as well. The minimum block size is one byte.
byte bSDRegCSDWriteBlPartial
'' Shows the size of the protected area.
'' The actual area =(SIZE_OF_PROTECTED_AREA) * MULT * BLOCK_LEN.
long lSDProtectedAreaSize

'' NCR for read and write operations
byte bSDRwNCR


' ---------------------------------------------------------------------------------'
'           	Buffers for CSD, OCR, CID Registers, Status Block		   '
' ---------------------------------------------------------------------------------'
string sSDRegisterCSD$( SD_CSD_REGISTER_SIZE )
string sSDRegisterOCR$( SD_OCR_REGISTER_SIZE )
string sSDRegisterCID$( SD_CID_REGISTER_SIZE )

string sSDStatusBlock$( SD_STATUS_BLOCK_SIZE )


'' string data buffer for acceleration
string  sRawBlock$(READ_BL_LEN_MAX) '' 512

'' offset where the data within raw_block lies on the card/
long lRawBlockAddress

' ******************************************************************************** '
' ******************************************************************************** '
'                                  SPI FUNCTIONS                                   '
' ******************************************************************************** '
sub vSPIWrite(string spDataToWrite$)
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, spDataToWrite$, -8
end


sub bSPIRead(var byte bpvReadData)
  shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, bpvReadData, -8, 1
end


sub vSPIInit()
  byte blMask
  blMask = ( 1 shl SD_SPI_DATA_OUT_PIN ) bitor ( 1 shl SD_SPI_CLOCK_PIN )
  out SD_SPI_PORT, blMask, 255
  dir_pin SD_SPI_PORT, SD_SPI_DATA_IN_PIN,  	1		'' SPI Port, Input Pin as Input
  dir_pin SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, 	0		'' SPI Port, Output Pin as Output
  dir_pin SD_SPI_PORT, SD_SPI_CLOCK_PIN,    	0		'' SPI Port, Clock Pin as Clock
end
'                                                                                  '
' ******************************************************************************** '


' ******************************************************************************** '
' ******************************************************************************** '
'                                  SD FUNCTIONS                                    '
' ******************************************************************************** '
sub vSDInit()
  call vSPIInit()

#ifdef SD_I_SDCARD_DETECT
  dir_pin SD_I_PORT, SD_I_SDCARD_DETECT, 1
#endif

#ifdef SD_I_ERROR_INDICATOR
  dir_pin SD_I_PORT, SD_I_ERROR_INDICATOR, 0
#endif

#ifdef SD_I_WRITE_PROTECT
  dir_pin SD_I_PORT, SD_I_WRITE_PROTECT, 1
#endif

#ifdef SD_I_POWER
  dir_pin SD_I_PORT, SD_I_POWER, 0
#endif

#ifdef SD_I_CHIP_SELECT
  dir_pin SD_I_PORT, SD_I_CHIP_SELECT, 0
#endif

#cm
  call vSDDisable()
  call vSDPowerOff()
  call vSDErrOff()
  call vSDPowerOn()
#endcm

#ifdef TIGER_2
''  long R
''  R = set_sysvarn(SHIFT_SLOWER,0)
#endif
end


sub vSDPowerOn()
  byte blMask
  blMask = ( 1 shl SD_SPI_DATA_OUT_PIN ) bitor ( 1 shl SD_SPI_CLOCK_PIN )
  out SD_SPI_PORT, blMask, 255
#ifdef SD_XP_POWER
  xres(SD_XP_ADDRESS+SD_XP_POWER_OFFSET,SD_XP_POWER)
#endif
#ifdef SD_I_POWER
  out SD_I_PORT, 1 shl SD_I_POWER, 0
#endif
  wait_duration  50
end


sub vSDPowerOff()
  byte blMask
  blMask = ( 1 shl SD_SPI_DATA_OUT_PIN ) bitor ( 1 shl SD_SPI_CLOCK_PIN )
  out SD_SPI_PORT, blMask, 0
#ifdef SD_XP_POWER
  xset(SD_XP_ADDRESS+SD_XP_POWER_OFFSET,SD_XP_POWER)
#endif
#ifdef SD_I_POWER
  out SD_I_PORT, 1 shl SD_I_POWER, 255
#endif
  wait_duration  50
end


sub vSDEnable()
  '' set low to activate SD Card chip select
#ifdef SD_XP_CHIP_SELECT
  xres(SD_XP_ADDRESS,SD_XP_CHIP_SELECT)
#endif
#ifdef SD_I_CHIP_SELECT
  out SD_I_PORT, 1 shl SD_I_CHIP_SELECT, 0
#endif	
  wait_duration	2
end


sub vSDDisable()
  '' set high to deactivate SD Card chip select
#ifdef SD_XP_CHIP_SELECT
  xset(SD_XP_ADDRESS,SD_XP_CHIP_SELECT)
#endif
#ifdef SD_I_CHIP_SELECT
  out SD_I_PORT, 1 shl SD_I_CHIP_SELECT, 255
#endif	
  wait_duration	2
end


sub vSDErrOn()
  '' turn ERR LED on
#ifdef SD_XP_ERROR_INDICATOR
  xset(SD_XP_ADDRESS,SD_XP_ERROR_INDICATOR)
#endif
#ifdef SD_I_ERROR_INDICATOR
  out SD_I_PORT, 1 shl SD_I_ERROR_INDICATOR, 255
#endif
end


sub vSDErrOff()
  '' turn ERR LED off
#ifdef SD_XP_ERROR_INDICATOR
  xres(SD_XP_ADDRESS,SD_XP_ERROR_INDICATOR)
#endif
#ifdef SD_I_ERROR_INDICATOR
  out SD_I_PORT, 1 shl SD_I_ERROR_INDICATOR, 0
#endif
end


sub bSDError( var byte bpvResult; byte bpErrorCode )
  '' Turn off SD Card
  call vSDDisable()
  '' Turn ERR LED on
  call vSDErrOn()
  bpvResult = bpErrorCode
end


sub bSDIsCardDetected( var byte bpvResult )
  bpvResult = 0
#ifdef SD_XP_SDCARD_DETECT
  call vSDEnable()
  bpvResult = xpin(SD_XP_ADDRESS,SD_XP_SDCARD_DETECT)
  call vSDDisable()
#endif
#ifdef SD_I_SDCARD_DETECT
  call vSDEnable()
  in SD_I_PORT, bpvResult
  call vSDDisable()
  bpvResult=bit(bpvResult,SD_I_SDCARD_DETECT)
#endif
  if bpvResult<>0 then
    call bSDError( bpvResult, ERROR_CARD_DETECT )
  endif
end


sub bSDIsCardWriteProtected( var byte bpvResult )
  bpvResult = 0
#ifdef SD_XP_WRITE_PROTECT
  call vSDEnable()
  bpvResult = xpin(SD_XP_ADDRESS,SD_XP_WRITE_PROTECT)
  call vSDDisable()
#endif
#ifdef SD_I_WRITE_PROTECT
  call vSDEnable()
  in SD_I_PORT, bpvResult
  call vSDDisable()
  bpvResult=bit(bpvResult,SD_I_WRITE_PROTECT)
#endif	
  if bpvResult<>0 then
    call bSDError( bpvResult, ERROR_CARD_WP )
  endif
end


sub bSDSetFramingBits( var byte bpvCmd; byte bpCmdId )
  byte blMask
  blMask = bitnot(10000000b)
  bpvCmd = bpCmdId
  bpvCmd = bpvCmd bitor 01000000b
  bpvCmd = bpvCmd bitand blMask
end


sub bSDWriteCmd( string spCommand$; var byte bpvResponse )
  long llTicks

  '' Turn off SD Card
  call vSDDisable()

  '' <-NCS-> - wait 8 clocks
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0ffh, -8

  '' Turn on SD Card
  call vSDEnable()

  '' Send the command to SD Card
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, spCommand$, -8

  ''<-NCR-> - wait 7 clocks for Tiger_2 and 8 clocks for Tiger_1
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, CLOCK_CYCLES, bSDRwNCR ''!!!

  llTicks = ticks()
  shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, bpvResponse, -8, 1
  while bpvResponse = 0ffh and diff_ticks( llTicks ) < SD_READ_RESPONSE_TIMEOUT
    shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, bpvResponse, -8, 1
  endwhile
end


'' ACMD41
sub bSDWriteOpCond( var byte bpvResult )
  byte blCount
  '' Wait for SD card to get ready
  blCount = 0
  while ( bpvResult bitand SD_R1_IN_IDLE_STATE ) <> 00h AND blCount < 50
    call bSDWriteCmd( SD_CM_SEND_OP_COND_COMMAND, bpvResult )
    wait_duration 50
    blCount = blCount + 1
  endwhile

  if ( bpvResult bitand SD_R1_IN_IDLE_STATE ) <> 00h then
    call bSDError( bpvResult, ERROR_SEND_OP_COND )
    return
  endif
end


'' ACMD13
sub bSDReadStatusBlock( var string spvStatus$; var byte bpvResult )
  byte blCurrentByte
  word wlIndex

  call bSDWriteCmd( SD_SEND_STATUS_COMMAND, bpvResult )

  if bpvResult <> 00h then
    '' Following any command, the SD Card needs 8 clocks to finish up its work.
    '' (from SanDisk SD Card Product Manual v1.9 section 5.1.8)	
    '' <-NEC-> - wait 8 clocks
    shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

    call bSDError( bpvResult, ERROR_ACMD13 )
    return
  else
    '' Now wait for the "Start Block" token	(0xFE)
    '' (see SanDisk SD Card Product Manual v1.9 section 5.2.4. Data Tokens)
    wlIndex = 0
    call bSPIRead( bpvResult )
    while bpvResult <> 0FEh AND wlIndex < 511
      call bSPIRead( bpvResult )
      wlIndex = wlIndex + 1
    endwhile
    if bpvResult <> 0FEh AND wlIndex >= 511 then
      call bSDError( bpvResult, ERROR_ACMD13 )
      return
    endif

    '' Read data bytes block
    spvStatus$ = ""
    for wlIndex = 0 to 63
      call bSPIRead( blCurrentByte )
      spvStatus$ = ntos$( spvStatus$, wlIndex, 1, blCurrentByte )
    next

    '' Following any command, the SD Card needs 8 clocks to finish up its work.
    '' (from SanDisk SD Card Product Manual v1.9 section 5.1.8)	
    '' <-NEC-> - wait 8 clocks
    shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

    '' Turn off SD Card
    call vSDDisable()

  endif
  bpvResult = 0	
end


'' This routine is incomplete. It reads only the size of the protected area,
'' but ignores other fields of the status block.
sub vSDParseStatusBlock( string spStatus$ )
  byte blCurrentByte, blIndex
  long llTemp
  lSDProtectedAreaSize = 00h		
  for blIndex = 0 to 63
    blCurrentByte = nfroms( spStatus$, blIndex, 1 )
    switchi blIndex
    case 4:
      llTemp = blCurrentByte
      llTemp = llTemp SHL 24			
      lSDProtectedAreaSize = lSDProtectedAreaSize bitor llTemp
    case 5:
      llTemp = blCurrentByte
      llTemp = llTemp SHL 16	
      lSDProtectedAreaSize = lSDProtectedAreaSize bitor llTemp
    case 6:
      llTemp = blCurrentByte
      llTemp = llTemp SHL 8			
      lSDProtectedAreaSize = lSDProtectedAreaSize bitor llTemp
    case 7:
      lSDProtectedAreaSize = lSDProtectedAreaSize bitor blCurrentByte
    endswitch
  next
end


'************************* ACMD13 = SD_STATUS ***********************************
sub bSDReadStatus( var byte bpvResult )
  call bSDWriteCmd( SD_APP_CMD_COMMAND, bpvResult )
  call bSDReadStatusBlock( sSDStatusBlock$, bpvResult )
  call vSDParseStatusBlock( sSDStatusBlock$ )
end


'' ******************* SET_BLOCKLEN *********************************************
sub bSDWriteBlockLen( word wpMaxBlockLen; var byte bpvResult )
  byte blCmd, blIndex
  string slCmdBlock$(CMD_BLOCK_SIZE)
  bpvResult = 0FFh
  ''word maxBlockLen
  ''maxBlockLen = maxBlockLen_Const

  '' ****** Creating a command with arguments **********
  slCmdBlock$ = SD_SET_BLOCK_LEN_COMMAND
  blCmd = wpMaxBlockLen SHR 8
  slCmdBlock$ = ntos$( slCmdBlock$, 3, 1, blCmd)
  blCmd = wpMaxBlockLen
  slCmdBlock$ = ntos$( slCmdBlock$, 4, 1, blCmd)

  '' ***************************************************
  blIndex=0
  while bpvResult <> 00h AND blIndex < 50
    call bSDWriteCmd( slCmdBlock$, bpvResult )
    blIndex = blIndex + 1
  endwhile


#ifdef DEBUG
  print_using  #SER, "response after SET_BLOCKLEN: ";  bpvResult
#endif
  if bpvResult <> 00h then
    call bSDError( bpvResult, ERROR_SET_BLOCKLEN )
  endif
end


sub bSDInitCard( var byte bpvResult )
  long llCount

  '' SD card reset
  llCount = 0
  call bSDWriteCmd( SD_CM_GO_IDLE_STATE_COMMAND, bpvResult )
  while bpvResult <> 01h and llCount < 50
    call bSDWriteCmd( SD_CM_GO_IDLE_STATE_COMMAND, bpvResult )
    llCount = llCount + 1
  endwhile

  if bpvResult <> 01h then
    call bSDError( bpvResult, ERROR_GO_TO_IDLE_STATE )
    return
  endif

#cm
  bpvResult = 0
  while ( bpvResult bitand SD_R1_IN_IDLE_STATE ) <> 00h AND llCount < 50
    call bSDWriteCmd( SD_CMD1_SEND_OP_COND_COMMAND, bpvResult )
    llCount = llCount + 1
  endwhile

  if ( bpvResult bitand SD_R1_IN_IDLE_STATE ) <> 00h then
    call bSDError( bpvResult, ERROR_GO_TO_IDLE_STATE )
    return
  endif

  call bSDWriteCmd( SD_CMD55_APP_CMD_COMMAND, bpvResult )
#endcm

  '' Wait for SD card to get ready
  call bSDWriteOpCond( bpvResult )
  if bpvResult <> 0 then
    return '' Time out
  endif
	
  '' Send MMC CMD16(SET_BLOCKLEN) to set the block length
  call bSDWriteBlockLen(0200h, bpvResult )	
end


'' ***************************************************************************** ''
'' 				Read CSD Register				 ''
'' ***************************************************************************** ''
sub bSDReadCSDRegister( var byte bpvResult )
  call bSDReadRegister( SD_SEND_CSD_COMMAND, sSDRegisterCSD$, bpvResult )
  if bpvResult <> 00h then
    call bSDInitCard(bpvResult)
    if bpvResult = 0 then
      call bSDReadRegister( SD_SEND_CSD_COMMAND, sSDRegisterCSD$, bpvResult )
    endif
  endif

  if bpvResult <> 00h then
    call bSDError( bpvResult, ERROR_GET_CSD )
    return	
  else
    call vSDParseCSDFields ( sSDRegisterCSD$ )
  endif
end


'' ***************************************************************************** ''
'' 				Read OCR Register				 ''
'' ***************************************************************************** ''
sub bSDReadOCRRegister( var byte bpvResult )
  call bSDReadRegister( SD_READ_OCR_COMMAND, sSDRegisterOCR$, bpvResult )
  if bpvResult <> 00h then
    call bSDInitCard(bpvResult)
    if bpvResult = 0 then
      call bSDReadRegister( SD_READ_OCR_COMMAND, sSDRegisterOCR$, bpvResult )
    endif
  endif

  if bpvResult <> 00h then
    call bSDError( bpvResult, ERROR_GET_OCR )
    return	
  else
    call vSDParseOCRFields ( sSDRegisterOCR$ )
  endif
end


'' ***************************************************************************** ''
'' 				Read CID Register				 ''
'' ***************************************************************************** ''
sub bSDReadCIDRegister( var byte bpvResult )
  call bSDReadRegister( SD_SEND_CID_COMMAND, sSDRegisterCID$, bpvResult )
  if bpvResult <> 00h then
    call bSDInitCard(bpvResult)
    if bpvResult = 0 then
      call bSDReadRegister( SD_SEND_CID_COMMAND, sSDRegisterCID$, bpvResult )
    endif
  endif

  if bpvResult <> 00h then
    call bSDError( bpvResult, ERROR_GET_CID )
    return
  else
    call vSDParseCIDFields ( sSDRegisterCID$ )
  endif
end


sub bSDReadConfiguration( var byte bpvResult )
  bpvResult = 0
  call bSDReadCSDRegister( bpvResult )
  call bSDReadOCRRegister( bpvResult )
  call bSDReadCIDRegister( bpvResult )
end


'' ***************************************************************************** ''
'' *  				Reset SD Card                                              * ''	
'' ***************************************************************************** ''
sub bSDResetCard(var byte bpvResult)
  string slDummyData$(16)
  bpvResult = 0FFh

  call vSDEnable()
  call vSDPowerOff()  ' !!!
  call vSDPowerOn()   ' !!!
''  call vSDDisable()

  '' We need to give SD Card about a hundred clock cycles to boot up
  '' 16*8 = 128 clocks
  slDummyData$ = fill$( "<0FFh>", 16)

  '' write dummy data to pump clock signal line
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, slDummyData$, 8

  call bSDInitCard( bpvResult )

#cm
  call vSDDisable()
  call vSDPowerOff()  ' !!!
  wait_duration 500
  call vSDPowerOn()   ' !!!
  wait_duration 2000
  '' We need to give SD Card about a hundred clock cycles to boot up
  '' 16*8 = 128 clocks
  slDummyData$ = fill$( "<0FFh>", 16)

  '' write dummy data to pump clock signal line
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, slDummyData$, 8
  call vSDEnable()

  call bSDInitCard( bpvResult )
''''  call vSDDisable()
#endcm
end ' end bSDResetCard


'' ***************************************************************************************************''
'' In SPI mode only the OCR, CSD and CID registers are accessible. Their format is identical to the   ''
'' format in the SD mode. However, a few fields are irrelevant in SPI mode.                           ''
'' ***************************************************************************************************''
sub bSDReadRegister( string spSDCommand$; var string spvData$; var byte bpvResult )

  '' In addition, the SD Card host can read the cards CID register using the READ_CID SD Card command. The CID
  '' register is programmed during the SD Card testing and formatting procedure, on the manufacturing floor. The SD
  '' Card host can only read this register and not write to it.
  byte blCounter, blLength, blReadByte

  bpvResult = 0FFh

  call bSDWriteCmd( spSDCommand$, bpvResult )

  if bpvResult <> 0 then
    call bSPIRead( bpvResult ) '' Read Error Token
  else
    bpvResult = 0
    if spSDCommand$ <> SD_READ_OCR_COMMAND then
      '' Now wait for the "Start Block" token	(0xFE)
      '' (see SanDisk SD Card Product Manual v1.9 section 5.2.4. Data Tokens)
      blCounter = 0
      call bSPIRead( bpvResult )
      while bpvResult <> 0FEh AND blCounter < 100
	call bSPIRead( bpvResult )
	blCounter = blCounter + 1	
      endwhile

      if blCounter = 100 AND  bpvResult <> 0FEh then
	return				
      endif
    endif

    blLength = max_len( spvData$ )
    for blCounter = 0 to blLength - 1
      call bSPIRead( blReadByte )
      spvData$ = ntos$( spvData$, blCounter, 1, blReadByte )
    next
  endif

  '' Following any command, the SD Card needs 8 clocks to finish up its work.
  '' (from SanDisk SD Card Product Manual v1.9 section 5.1.8)	
  '' <-NEC-> - wait 8 clocks
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

  '' Turn off SD Card
  call vSDDisable()

  bpvResult = 0
end


'' **************************************************************************** ''
'' *                           Parse CSD register data                        * ''
'' **************************************************************************** ''
sub vSDParseCSDFields ( string spData$ )
  byte blIndex, blLength, blCurrentByte, blBuffer
  word wlBuffer
  wlBuffer = 00h		
  blBuffer = 00h

  '' Clean sector
  bSDRegCSDReadBlLen      = 00h
  bSDRegCSDReadBlPartial  = 00h
  wSDRegCSDCsize          = 00h
  bSDRegCSDCsizeMult      = 00h
  bSDRegCSDSectorSize     = 00h
  bSDRegCSDWpGrpSize      = 00h
  bSDRegCSDWriteBlLen     = 00h
  bSDRegCSDWriteBlPartial = 00h
  ''  bSDRegCSDTaac            = 00h
  ''  bSDRegCSDNsac            = 00h

  blLength = len( spData$ )
  for blIndex = 0 to blLength - 1	
    blCurrentByte = nfroms( spData$, blIndex, 1 )
    switchi blIndex
    case 0:
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
      bSDRegCSDReadBlLen = blCurrentByte bitand 0Fh
    case 6:
      bSDRegCSDReadBlPartial = BIT_MIRR( (blCurrentByte bitand 80h ), 8 )
      wlBuffer = blCurrentByte bitand 03h
      wSDRegCSDCsize = wlBuffer SHL 10	
      wlBuffer = 00h
    case 7:
      wlBuffer = blCurrentByte
      wlBuffer = wlBuffer SHL 2
      wSDRegCSDCsize = wSDRegCSDCsize bitor wlBuffer		
      wlBuffer = 00h
    case 8:
      wlBuffer = blCurrentByte bitand 0C0h
      wlBuffer = wlBuffer SHR 6
      wSDRegCSDCsize = wSDRegCSDCsize bitor wlBuffer
      wlBuffer = 00h
    case 9:
      bSDRegCSDCsizeMult = blCurrentByte bitand 03h
      bSDRegCSDCsizeMult = bSDRegCSDCsizeMult SHL 1
    case 10:
      blBuffer = (blCurrentByte bitand 80h) SHR 7
      bSDRegCSDCsizeMult = bSDRegCSDCsizeMult bitor blBuffer

      blBuffer = (blCurrentByte bitand 3Fh) SHL 1
      bSDRegCSDSectorSize = blBuffer
      blBuffer = 00h
    case 11:
      blBuffer = (blCurrentByte bitand 80h) SHR 7
      bSDRegCSDSectorSize = bSDRegCSDSectorSize bitor blBuffer
      blBuffer = 00h

      bSDRegCSDWpGrpSize = blCurrentByte bitand 7Fh
    case 12:
      blBuffer = (blCurrentByte bitand 03h) SHL 2
      bSDRegCSDWriteBlLen = blBuffer
      blBuffer = 00h
    case 13:
      blBuffer = (blCurrentByte bitand 0C0h) SHR 6
      bSDRegCSDWriteBlLen = bSDRegCSDWriteBlLen bitor blBuffer
      bSDRegCSDWriteBlPartial = (blCurrentByte bitand 020h) SHR 5
    case 14:
    case 15:
    endswitch
  next
end


'' **************************************************************************** ''
'' *                           Parse CID register data                        * ''
'' **************************************************************************** ''
sub vSDParseCIDFields ( string spData$ )
end


'' **************************************************************************** ''
'' *                           Parse OCR register data                        * ''
'' **************************************************************************** ''
sub vSDParseOCRFields ( string spData$ )
end


sub lSDGetBlockLenVar( var long lpvBytesInBlock )
  lpvBytesInBlock = lSDBlockLen
end


sub vSDSetBlockLenVar( long lpBytesInBlock )
  lSDBlockLen = lpBytesInBlock
end


sub vSDSetRwNCRVar( byte bpRwNCR )
  bSDRwNCR = bpRwNCR
end


'' The memory capacity of the card is computed from the entries C_SIZE, C_SIZE_MULT and
'' READ_BL_LEN as follows:
'' memory capacity = BLOCKNR * BLOCK_LEN
'' Where:
'' BLOCKNR = (C_SIZE+1) * MULT
'' MULT = 2C_SIZE_MULT+2 (C_SIZE_MULT < 8)
'' BLOCK_LEN = 2READ_BL_LEN (READ_BL_LEN < 12)
'' Therefore, the maximum capacity which can be coded is 4096*512*2048 = 4 GBytes.
'' Example: A four MByte card with BLOCK_LEN = 512 can be coded with C_SIZE_MULT = 0 and C_SIZE = 2047.
sub lSDCalcCardSize( var long lpvSDCardSize )
  long llCurProtectedAreaSize, llMult, llBlockNr, llBlockLen

  llMult = 1 SHL ( bSDRegCSDCsizeMult + 2 )
  llBlockNr = ( wSDRegCSDCsize + 1 ) * llMult
  llBlockLen = 1 SHL bSDRegCSDReadBlLen

  '' The actual protected area size =(SIZE_OF_PROTECTED_AREA) * MULT * BLOCK_LEN.
  llCurProtectedAreaSize = lSDProtectedAreaSize * llMult * llBlockLen
  lpvSDCardSize = llBlockNr * llBlockLen + llCurProtectedAreaSize
end


sub bSDAbortWithError( var byte bpvResult; byte bpErrorCode )
  byte blResponse
  '' ABORT: invalid response for read single command
  '' Read Error Token
  shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1

  '' Following any command, the SD Card needs 8 clocks to finish up its work.
  '' (from SanDisk SD Card Product Manual v1.9 section 5.1.8)	
  '' <-NEC-> - wait 8 clocks
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

  call bSDError( bpvResult, bpErrorCode )
end


'' **************************************************************************** ''
'' *                           READ SINGLE BLOCK                              * ''
'' **************************************************************************** ''
'' lpAddress - The offset from which to read.
'' lpSize - The number of bytes to read.
sub bSDReadBlock(long lpAddress; long lpSize; var string spvReadData$; var byte bpvResult )

  byte blCmd, blResponse
  word wlCRC
  long llBlockAddress, llBlockOffset, llReadLength, llOffset, llLength, llIndex, llBufOffset
  string slCmdBlock$(CMD_BLOCK_SIZE)

  llOffset = lpAddress
  llLength = lpSize

  if (llLength > max_len(spvReadData$)) then
    llLength = max_len(spvReadData$)
  endif		

  '' *** Creating a READ_SINGLE_BLOCK command with arguments ***
  call bSDSetFramingBits( blCmd, SD_CM_READ_SINGLE_BLOCK )
  slCmdBlock$ = fill$( "<00h>", CMD_BLOCK_SIZE )
  slCmdBlock$ = ntos$( slCmdBlock$, 0, 1, blCmd)
  slCmdBlock$ = ntos$( slCmdBlock$, 1, -4, 0ffffffffh )
  slCmdBlock$ = ntos$( slCmdBlock$, 5, 1, 0ffh)
  '' ***************************************************

  llBufOffset = 0
  while llLength > 0
    '' determine byte count to read at once/
    llBlockAddress = llOffset bitand 0fffffe00h
    llBlockOffset  = llOffset bitand 001ffh
    '' read up to block border
    llReadLength = lSDBlockLen - llBlockOffset '' READ_BL_LEN
    if  llReadLength > llLength then
      llReadLength = llLength
    endif

    slCmdBlock$ = ntos$( slCmdBlock$, 1, -4, llBlockAddress)

    call bSDWriteCmd( slCmdBlock$, blResponse )

    if blResponse = 0ffh then
      call bSDAbortWithError( bpvResult, ERROR_READ_SINGLE_BLOCK )
      return
    else
      '' Now wait for the "Start Block" token	(0xFE)
      '' (see SanDisk SD Card Product Manual v1.9 section 5.2.4. Data Tokens)
      llIndex = 0
      while blResponse <> 0FEh AND llIndex < 511
	''call bSPIRead(blResponse)
	shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
	llIndex = llIndex+1
      endwhile
      if blResponse <> 0FEh AND llIndex>=511 then
	call bSDError( bpvResult, ERROR_READ_SINGLE_BLOCK )
	return
      endif

      '' Read bytes block
      sRawBlock$ = fill$("<0>",512)	
      shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, sRawBlock$, -8, 1
      spvReadData$ = stos$( spvReadData$, llBufOffset, sRawBlock$, llBlockOffset, llReadLength )
      llBufOffset = llBufOffset + llReadLength

      '' Read CRC bytes
      shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, wlCRC, -16, 1
    endif

    '' Following any command, the SD Card needs 8 clocks to finish up its work.
    '' (from SanDisk SD Card Product Manual v1.9 section 5.1.8)	
    '' <-NEC-> - wait 8 clocks
    shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

    '' Turn off SD Card
    call vSDDisable()

    llLength = llLength - llReadLength
    llOffset = llOffset + llReadLength

  endwhile

  bpvResult = 0	
end ''READ SINGLE BLOCK


'' **************************************************************************** ''
'' *                           READ MULTIPLE BLOCK                            * ''
'' **************************************************************************** ''
sub bSDReadMultipleBlock(long lpAddress; long lpSize; var string spvReadData$; var byte bpvResult)
  byte blCmd, blResponse
  word wlCRC, wlCurrentWriteLen
  long llBlockAddress, llBlockOffset, llReadLength, llOffset, llLength, llBufIndex, llIndex
  string slCmdBlock$(CMD_BLOCK_SIZE)

  llOffset = lpAddress
  llLength = lpSize
  llBufIndex = 0
  wlCurrentWriteLen = 0

  if (llLength > max_len(spvReadData$))	then
    llLength = max_len(spvReadData$)
  endif		

  '' determine byte count to read at once/
  llBlockAddress = llOffset bitand 0fffffe00h
  llBlockOffset  = llOffset bitand 001ffh

  '' read up to block border
  llReadLength = llLength + llBlockOffset

  '' *** Creating a READ_MULTIPLE_BLOCK command with arguments ***
  slCmdBlock$ = fill$( "<00h>", CMD_BLOCK_SIZE )
  call bSDSetFramingBits( blCmd, SD_CM_READ_MULTIPLE_BLOCK )
  slCmdBlock$ = ntos$( slCmdBlock$, 0, 1, blCmd)
  slCmdBlock$ = ntos$( slCmdBlock$, 1, -4, llBlockAddress)
  slCmdBlock$ = ntos$( slCmdBlock$, 5, 1, 0FFh)
  '' ***************************************************

  '' Turn on SD Card
  call vSDEnable()

  '' Send the command to SD Card
  ''call vSPIWrite(slCmdBlock$)
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, slCmdBlock$, -8

  ''<-NCR-> - wait 7 clocks for Tiger_2 and 8 clocks for Tiger_1
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, bSDRwNCR ''!!!

  '' Wait for the response
  ''call bSPIRead(blResponse)
  shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1

  llIndex=0
  while blResponse <> 0 and llIndex < 8
    ''call bSPIRead(blResponse)
    shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
    llIndex = llIndex + 1		
  endwhile

  if blResponse <> 0 then
    call bSDAbortWithError( bpvResult, ERROR_READ_MULTIPLE_BLOCK )
    return
  else
#cm
'' Now wait for the "Start Block" token	(0xFE)
'' (see SanDisk SD Card Product Manual v1.9 section 5.2.4. Data Tokens)
blResponse = 0FFh
llIndex = 0
while blResponse <> 0FEh AND llIndex<511
''call bSPIRead(blResponse)
shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
llIndex = llIndex+1
endwhile
if blResponse <> 0FEh AND llIndex>=511 then
bStatus = ERROR_READ_MULTIPLE_BLOCK
return
endif

for llIndex=0 to llReadLength -1
if llIndex<>0 AND mod( llIndex, lSDBlockLen) = 0 then
blCmd = 0FFh
while blCmd <> 0FEh
call bSPIRead(blCmd)    		
endwhile
endif		
call bSPIRead(blCmd)    		
if llIndex >= llBlockOffset then
spvReadData$ = ntos$( spvReadData$, llBufIndex, 1, blCmd)
llBufIndex = llBufIndex+1
endif
next
#endcm

    for llIndex=0 to llReadLength -1 STEP 512
      '' Now wait for the "Start Block" token	(0xFE)
      '' (see SanDisk SD Card Product Manual v1.9 section 5.2.4. Data Tokens)
      blResponse = 0FFh
      llIndex = 0
      while blResponse <> 0FEh AND llIndex<511
	''call bSPIRead(blResponse)
	shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
	llIndex = llIndex+1
      endwhile
      if blResponse <> 0FEh AND llIndex>=511 then
	call bSDError( bpvResult, ERROR_READ_MULTIPLE_BLOCK )
	return
      endif

      '' Read bytes block
      sRawBlock$ = fill$("<0>",512)	
      shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, sRawBlock$, -8, 1

      if llIndex = 0 then
	if(512-llBlockOffset) >= llLength then
	  llBufIndex = llLength		
	else
	  llBufIndex = 512-llBlockOffset
	endif		
	spvReadData$ = MID$(sRawBlock$, llBlockOffset, llBufIndex)
      else
	if(llLength - llBlockOffset) >= 512 then
	  wlCurrentWriteLen = 512
	  llBlockOffset = llBlockOffset + 512
	else
	  wlCurrentWriteLen = llLength - llBlockOffset					
	endif			
	spvReadData$ = stos$(spvReadData$, llBufIndex, sRawBlock$, 0, wlCurrentWriteLen)
	''source_indexer = source_indexer + write_length		
      endif	

      '' Read CRC bytes
      shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, wlCRC, -16, 1
    next	

    '' *** Creating a STOP_TRANSMISSION command with arguments ***
    slCmdBlock$ = fill$( "<00h>", CMD_BLOCK_SIZE )
    call bSDSetFramingBits( blCmd, SD_CM_STOP_TRANSMISSION )
    slCmdBlock$ = ntos$( slCmdBlock$, 0, 1, blCmd)
    slCmdBlock$ = ntos$( slCmdBlock$, 5, 1, 0FFh)
    '' ***************************************************
    '' Send the command to SD Card
    ''call vSPIWrite(slCmdBlock$)
    shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, slCmdBlock$, -8

    ''<-NCR-> - wait 7 clocks for Tiger_2 and 8 clocks for Tiger_1
    shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, bSDRwNCR ''!!!

    '' Wait for the response
    ''call bSPIRead(blResponse)
    shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1

    llIndex=0
    while blResponse <> 0 and llIndex < 8
      ''call bSPIRead(blResponse)
      shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
      llIndex = llIndex + 1		
    endwhile


    if blResponse <> 0 then
      call bSDAbortWithError( bpvResult, ERROR_STOP_TRANSMISSION )
      return
    endif
  endif

  '' Following any command, the SD Card needs 8 clocks to finish up its work.
  '' (from SanDisk SD Card Product Manual v1.9 section 5.1.8)	
  '' <-NEC-> - wait 8 clocks
  shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

  '' wait while card is busy
  while blResponse <> 0FFh
    ''call bSPIRead(blResponse)
    shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
  endwhile

  '' Turn off SD Card
  call vSDDisable()

  bpvResult = 0	

end ''READ MULTIPLE BLOCK


'' **************************************************************************** ''
'' *                            Write Single Block                            * ''
'' **************************************************************************** ''
sub bSDWriteBlock(long lpAddress; long lpSize; var string spvDataToWrite$;  var byte bpvResult)

  byte blCmd, blResponse
  long llBlockAddress, llBlockOffset, llWriteLength, llOffset, llLength, llIndex, llSourceIndexer
  string slCmdBlock$(CMD_BLOCK_SIZE)

  bpvResult = 0 '' !!!

  call bSDIsCardWriteProtected( bpvResult )
  if bpvResult <> 0 then
    return
  endif

  llOffset = lpAddress
  llLength = lpSize
  llSourceIndexer = 0
  lRawBlockAddress = 0000h
  sRawBlock$ = fill$( "<00h>", READ_BL_LEN_MAX )

  '' ****** Creating a "write" command with arguments **********
  call bSDSetFramingBits( blCmd, SD_CM_WRITE_BLOCK )
  slCmdBlock$ = fill$( "<00h>", CMD_BLOCK_SIZE )
  slCmdBlock$ = ntos$( slCmdBlock$, 0, 1, blCmd )
  slCmdBlock$ = ntos$( slCmdBlock$, 1, -4, 0ffffffffh )
  slCmdBlock$ = ntos$( slCmdBlock$, 5, 1, 0FFh)
  '' *************************************************

  while llLength > 0
    '' determine byte count to read at once
    llBlockAddress = llOffset bitand 0fffffe00h
    llBlockOffset  = llOffset bitand 001ffh
    '' read up to block border
    llWriteLength = 512 - llBlockOffset
    if llWriteLength > llLength   then
      llWriteLength = llLength
    endif

    '' Merge the data to write with the content of the block.
    ''  Use the cached block if available.
    if llBlockAddress <> lRawBlockAddress then
      if llBlockOffset <> 0 OR llWriteLength < 512  then
	bpvResult = ERROR_WRITE_SINGLE_BLOCK	
#ifdef READ_MULTIPLE_BLOCK
	call bSDReadMultipleBlock(llBlockAddress, READ_BL_LEN_MAX, sRawBlock$, bpvResult)
#else
	call bSDReadBlock(llBlockAddress, READ_BL_LEN_MAX, sRawBlock$, bpvResult)
#endif
	if bpvResult <> 0 then
	  return	
	endif			
      endif
    endif

    lRawBlockAddress = llBlockAddress
    sRawBlock$ = stos$(sRawBlock$, llBlockOffset, spvDataToWrite$, llSourceIndexer, llWriteLength)
    llSourceIndexer = llSourceIndexer + llWriteLength

    slCmdBlock$ = ntos$( slCmdBlock$, 1, -4, llBlockAddress )

    '' Initialize return vallue by default
    blResponse = 0FFh

        '' Send command to the board
    call bSDWriteCmd( slCmdBlock$, blResponse )

    if blResponse <> 0h then
      call bSDAbortWithError( bpvResult, ERROR_WRITE_SINGLE_BLOCK )
      return
    else
      '' <-NWR-> - wait 8x clocks
      shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, CLOCK_CYCLES, 8

      '' Write Start Byte FEh
      shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FEh, -8 ''7Fh, 8
			
      '' Write bytes block
      '' call vSPIWrite(sRawBlock$)
      shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, sRawBlock$, -8

      '' Write dummy crc16
      shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, -8
      shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, -8
			
      '' Response
      shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
			
      if (blResponse and 0Fh) <> 05h then
        call bSDAbortWithError( bpvResult, ERROR_WRITE_SINGLE_BLOCK )
        return
      endif  	

      '' Card is busy
      call bWaitForWriteFinish( blResponse )

      if blResponse <> 0h  then
        call bSDAbortWithError( bpvResult, ERROR_WRITE_SINGLE_BLOCK )
        return
      endif
			
      '' Turn off SD Card
      call vSDDisable ()		
			
      '' Following any command, the SD Card needs 8 clocks to finish up its work.
      '' <-NEC-> - wait 8 clocks
      shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8			
			
    endif				

    llLength = llLength - llWriteLength
    llOffset = llOffset + llWriteLength

  endwhile

end '' bSDWriteBlock


'***************** MMC wait for write finish ********************'
'															     '
' Repeatedly reads the MMC until we get a non-zero value (after  '
' a zero value) indicating the write has finished and card is no '
' longer busy.                                                   '
'                                                                '
'****************************************************************'
sub bWaitForWriteFinish(var byte bpvResult)
  byte blResponse
  word wlCount

  blResponse=0
  '' The delay is set to maximum considering the longest data block length to handle
  wlCount = 0FFFFh
	
  while (blResponse = 0) and (wlCount<>0)
    shift_in SD_SPI_PORT, SD_SPI_DATA_IN_PIN, SD_SPI_CLOCK_PIN, blResponse, -8, 1
    wlCount = wlCount - 1
  endwhile
	
  if wlCount = 0 then
    bpvResult = 1 '' Failure, loop was exited due to timeout  	
  else
    bpvResult = 0 '' Normal, loop was exited before timeout
  endif
end

#cm
'' **************************************************************************** ''
'' *                            Write Multiple Block                          * ''
'' **************************************************************************** ''
sub SD_Write_Multiple_Block(long lpAddress; long lpSize; var string spvDataToWrite$;  var byte bStatus)

byte blCmd, blResponse, bTMP, bFirstBlockFromBuffer
long llBlockAddress, llBlockOffset, llWriteLength, llOffset, llLength, llIndex, llSourceIndexer, lCurrentOffsetInSource

bStatus = 0 '' !!!

llOffset = lpAddress
llLength = lpSize
llSourceIndexer = 0
lRawBlockAddress = 0000h
sRawBlock$ = fill$( "<00h>", READ_BL_LEN_MAX )
bFirstBlockFromBuffer = 0
lCurrentOffsetInSource = 0

''   if(get_pin_locked())
''        return 0;

while llLength > 0

'' determine byte count to read at once
llBlockAddress = llOffset bitand 0fffffe00h
llBlockOffset  = llOffset bitand 001ffh
'' read up to block border
llWriteLength = 512 - llBlockOffset
if llWriteLength > llLength   then
llWriteLength = llLength
endif

'' Merge the data to write with the content of the block.
'' Use the cached block if available.
'' Only for the first block
if llBlockAddress <> lRawBlockAddress then
if llBlockOffset <> 0 OR llWriteLength < 512  then
bStatus = ERROR_WRITE_BLOCK	
call bSDReadBlock(llBlockAddress, READ_BL_LEN_MAX, sRawBlock$, bStatus)
bFirstBlockFromBuffer = 1
if bStatus <> 0 then
return	
endif			
endif
endif

lRawBlockAddress = llBlockAddress
for llIndex=0 to llWriteLength-1  '' Attention please "-1"
bTMP = nfroms(spvDataToWrite$, llSourceIndexer+llIndex, 1)	
sRawBlock$ = ntos$(sRawBlock$, llBlockOffset+llIndex, 1, bTMP)		
next
llSourceIndexer = llWriteLength

'' ****** Creating a command with arguments **********
call bSDSetFramingBits( blCmd, SD_CM_WRITE_MULTI_BLOCK )
sCmdBlock$ = fill$( "<00h>", CMD_BLOCK_SIZE )
sCmdBlock$ = ntos$( sCmdBlock$, 0, 1, blCmd )

blCmd = llBlockAddress SHR 24
blCmd = BIT_MIRR(blCmd, 8)
sCmdBlock$ = ntos$( sCmdBlock$, 1, 1, blCmd )

blCmd = llBlockAddress SHR 16
blCmd = BIT_MIRR(blCmd, 8)
sCmdBlock$ = ntos$( sCmdBlock$, 2, 1, blCmd )

blCmd = llBlockAddress SHR 8
blCmd = BIT_MIRR(blCmd, 8)
sCmdBlock$ = ntos$( sCmdBlock$, 3, 1, blCmd)

blCmd = llBlockAddress SHR 0
blCmd = BIT_MIRR(blCmd, 8)
sCmdBlock$ = ntos$( sCmdBlock$, 4, 1, blCmd)

sCmdBlock$ = ntos$( sCmdBlock$, 5, 1, 0FFh)
'' **************************************************

'' Initialize return vallue by default
blResponse = 0FFh
'' Turn on SD Card
call vSDEnable()
'' <-NCS-> - wait 8 clocks
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8	
'' Send Write Command to SD Card
call vSPIWrite(sCmdBlock$)
''<-NCR-> - wait 7 clocks for Tiger_2 and 8 clocks for Tiger_1
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, NCR ''!!!
'' Wait for the response
call bSPIRead(blResponse)
''<-NWR-> - wait 8x clocks
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

if blResponse <> 0 then
'' ABORT: invalid response for write multiple block command
call bSPIRead(blResponse)'' Read Error Token
'' Following any command, the SD Card needs 8 clocks to finish up its work.
'' (from SanDisk SD Card Product Manual v1.9 section 5.1.8)	
'' <-NEC-> - wait 8 clocks
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8
'' Turn off SD Card
call vSDDisable()
bStatus = ERROR_WRITE_MULTIPLE_BLOCK
return
else  	
'' Write Start Byte FCh for Multiple Block Write operation
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 3Fh, 8

if bFirstBlockFromBuffer = 1 then
'' Mirror bytes in sRawBlock$
for llIndex=0 to 511
bTMP = nfroms(sRawBlock$, llIndex, 1)
bTMP = BIT_MIRR(bTMP, 8)
sRawBlock$ = ntos$(sRawBlock$,llIndex, 1, bTMP)
next
'' Write byte block	from buffer		
call vSPIWrite(sRawBlock$)
else

endif

'' Write dummy crc16
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8


'' wait while card is busy
blResponse = 00h
while blResponse <> 0FFh
call bSPIRead(blResponse)
endwhile

endif

'' Following any command, the SD Card needs 8 clocks to finish up its work.
'' <-NEC-> - wait 8 clocks
shift_out SD_SPI_PORT, SD_SPI_DATA_OUT_PIN, SD_SPI_CLOCK_PIN, 0FFh, 8

'' Turn off SD Card
call vSDDisable ()

llLength = llLength - llWriteLength
llOffset = llOffset + llWriteLength

endwhile

end '' SD_Write_Multiple_Block
#endcm

#endif  '' __SD_CARD_SPI_I

